//
// webserver.cpp
//
// Simple HTTP server sample for sanos
//

#include "webserver.h"
#include "socket.h"

//============================================================================

static void __fprintf( int fd, const char *fmt, ... )
{
	static char buf[4096];
	va_list va;

	va_start( va, fmt );
	vsnprintf( buf, sizeof( buf ) - 1, fmt, va );
	va_end( va );
	send( fd, buf, strlen( buf ), 0 );
}

//============================================================================

void WebServer::ReadyToRun( void )
{
	int sock;
	Socket socket;

	sock = socket.Bind(PORT);

	socket.Listen(sock);
	printf( "HTTP server listening on port %d\n", PORT );

	while( 1 )
	{
		int s;

		s = socket.Accept(sock);
		if( s < 0 )
			break;

		Process ( s );
	}

	socket.Close(sock);
}

const char* WebServer::GetMimeType( const char *name )
{
	char *ext = strrchr( name, '.' );
	if( !ext )
		return NULL;
	if( strcmp( ext, ".html" ) == 0 )
		return "text/html";
	if( strcmp( ext, ".htm" ) == 0 )
		return "text/html";
	if( strcmp( ext, ".jpg" ) == 0 )
		return "image/jpeg";
	if( strcmp( ext, ".jpeg" ) == 0 )
		return "image/jpeg";
	if( strcmp( ext, ".gif" ) == 0 )
		return "image/gif";
	if( strcmp( ext, ".png" ) == 0 )
		return "image/png";
	if( strcmp( ext, ".css" ) == 0 )
		return "text/css";
	if( strcmp( ext, ".au" ) == 0 )
		return "audio/basic";
	if( strcmp( ext, ".wav" ) == 0 )
		return "audio/wav";
	if( strcmp( ext, ".avi" ) == 0 )
		return "video/x-msvideo";
	if( strcmp( ext, ".mpeg" ) == 0 )
		return "video/mpeg";
	if( strcmp( ext, ".mpg" ) == 0 )
		return "video/mpeg";
	if( strcmp( ext, ".mp3" ) == 0 )
		return "audio/mpeg";
	return NULL;
}

void WebServer::SendHeaders( int fd, int status, const char *title, const char *extra, const char *mime, int length, time_t date )
{
	time_t now;
	char timebuf[128];

	__fprintf( fd, "%s %d %s\r\n", PROTOCOL, status, title );
	__fprintf( fd, "Server: %s\r\n", SERVER );
	now = time( NULL );
	strftime( timebuf, sizeof( timebuf ), RFC1123FMT, gmtime( &now ) );
	__fprintf( fd, "Date: %s\r\n", timebuf );
	if( extra )
		__fprintf( fd, "%s\r\n", extra );
	if( mime )
		__fprintf( fd, "Content-Type: %s\r\n", mime );
	if( length >= 0 )
		__fprintf( fd, "Content-Length: %d\r\n", length );
	if( date != -1 )
	{
		strftime( timebuf, sizeof( timebuf ), RFC1123FMT, gmtime( &date ) );
		__fprintf( fd, "Last-Modified: %s\r\n", timebuf );
	}
	__fprintf( fd, "Connection: close\r\n" );
	__fprintf( fd, "\r\n" );
}

void WebServer::SendError( int fd, int status, const char *title, const char *extra, const char *text )
{
	SendHeaders( fd, status, title, extra, "text/html", -1, -1 );
	__fprintf( fd, "<HTML><HEAD><TITLE>%d %s</TITLE></HEAD>\r\n", status, title );
	__fprintf( fd, "<BODY><H4>%d %s</H4>\r\n", status, title );
	__fprintf( fd, "%s\r\n", text );
	__fprintf( fd, "</BODY></HTML>\r\n" );
}

void WebServer::SendFile( int fd, const char *path, struct stat *statbuf )
{
	char data[4096];
	int n;

	FILE *file = fopen( path, "r" );

	if( !file )
		SendError( fd, 403, "Forbidden", NULL, "Access denied." );
	else
	{
		int length = S_ISREG( statbuf->st_mode ) ? statbuf->st_size : -1;
		SendHeaders( fd, 200, "OK", NULL, GetMimeType( path ), length, statbuf->st_mtime );

		while( ( n = fread( data, 1, sizeof( data ), file ) ) > 0 )
			if( write( fd, data, sizeof( data ) ) != sizeof( data ) )
				perror( "write" );
		fclose( file );
	}
}

int WebServer::Process ( int fd )
{
	char buf[4096];
	char *method;
	char *path;
	char *protocol;
	struct stat statbuf;
	char pathbuf[4096];
	int len;
	FILE *f;
	char timebuf[32];
	struct tm *tm;
	DIR *dir;
	struct dirent *de;

	f = fdopen( fd, "a+" );
	if( !fgets( buf, sizeof( buf ), f ) ) // read single first line
		return -1;
	printf( "URL: %s", buf );

	method = strtok( buf, " " );
	path = strtok( NULL, " " );

	protocol = strtok( NULL, "\r" );
	if( !method || !path ||!protocol )
		return -1;

	fseek( f, 0, SEEK_CUR );	// Force change of stream direction

	if( strcasecmp( method, "GET" ) != 0 )
		SendError( fd, 501, "Not supported", NULL, "Method is not supported." );
	else if( stat( path, &statbuf ) < 0 )
		  SendError( fd, 404, "Not Found", NULL, "File not found." );

	else if( S_ISDIR( statbuf.st_mode ) )
	{
		len = strlen( path );

		if( len == 0 || path[len - 1] != '/' )
		{
			snprintf( pathbuf, sizeof( pathbuf ), "Location: %s/", path );

			SendError( fd, 302, "Found", pathbuf, "Directories must end with a slash." );
		}
		else
		{
			snprintf( pathbuf, sizeof( pathbuf ), "%sindex.html", path );

			if( stat( pathbuf, &statbuf ) >= 0 )
				SendFile( fd, pathbuf, &statbuf );
			else
			{
				SendHeaders( fd, 200, "OK", NULL, "text/html", -1, statbuf.st_mtime );
				__fprintf( fd, "<HTML><HEAD><TITLE>Index of %s</TITLE></HEAD>\r\n<BODY>", path );
				__fprintf( fd, "<H4>Index of %s</H4>\r\n<PRE>\n", path );

				__fprintf( fd, "Name                             Last Modified              Size\r\n" );
				__fprintf( fd, "<HR>\r\n" );
				if( len > 1 )
					__fprintf( fd, "<A HREF=\"..\">..</A>\r\n" );

				dir = opendir( path );

				while( ( de = readdir( dir ) ) != NULL )
				{
					/* ignore dot(.) and dotdot(..) */
					if( strcmp( de->d_name, "." ) == 0 || strcmp( de->d_name, ".." ) == 0 )
						continue;

					strcpy( pathbuf, path );
					strcat( pathbuf, de->d_name );
					stat( pathbuf, &statbuf );
					tm = gmtime( &statbuf.st_mtime );
					strftime( timebuf, sizeof( timebuf ), "%d-%b-%Y %H:%M:%S", tm );

					__fprintf( fd, "<A HREF=\"%s%s\">", de->d_name, S_ISDIR( statbuf.st_mode ) ? "/" : "" );
					__fprintf( fd, "%s%s", de->d_name, S_ISDIR( statbuf.st_mode ) ? "/</A>" : "</A> " );
					if( strlen( de->d_name ) < 32 )
						__fprintf( fd, "%*s", 32 - strlen( de->d_name ), "" );

					if( S_ISDIR( statbuf.st_mode ) )
						__fprintf( fd, "%s\r\n", timebuf );
					else
						__fprintf( fd, "%s %10d\r\n", timebuf, statbuf.st_size );
				}
				closedir( dir );

				__fprintf( fd, "</PRE>\r\n<HR>\r\n<ADDRESS>%s</ADDRESS>\r\n</BODY></HTML>\r\n", SERVER );
			}
		}
	}
	else
		SendFile( fd, path, &statbuf );

	fclose( f );
	return 0;
}

int main()
{
	WebServer server;
	server.ReadyToRun();
	return 0;
}

